# -*- coding: utf-8 -*-
"""
Created on Mon Aug  4 16:50:59 2014

@author: charleslelosq

Python Wrapper for the gcvspl.f spline function
It assumes that you already have the good path in your pathway
gcvspl.so should exist in this path...
"""

from . import _gcvspl as gcvspl
import numpy as np

def gcvspline(x,y,ese,VAL,**options):
    """
    
    c, wk, ier = gcvspline(x,y,f,**options)
    
    Function for using the wrapper of gcvspl.f
    Input:
    x (1D array) are the independent variables
    y (2D array as (len(x),1)) are the observations
    (we assume here that you want to use this 
    spline only on 1 dataset... see gcvspl.f if not)
    ese are the errors on y
    VAL (double) depends on the mode, see gcvspl.f for details (=ese**2 for mode3)
    options:
    NC is the number of output spline coefficients, NC >= len(y) 
        default: NC = len(y)
    splorder is the half order of the required B-splines. The values 
    splorder = 1,2,3,4 correspond to linear, cubic, quintic, and 
    heptic splines, respectively.
        default: splorder = 2 (cubic) 
    splmode is the Optimization mode switch:
        default: splmode = 2 (General Cross Validated)
                       splmode = 1: Prior given value for p in VAL
                                 (VAL.ge.ZERO). This is the fastest
                                 use of GCVSPL, since no iteration
                                 is performed in p.
                       splmode = 2: Generalized cross validation.
                       splmode = 3: True predicted mean-squared error,
                                 with prior given variance in VAL.
                       splmode = 4: Prior given number of degrees of
                                 freedom in VAL (ZERO.le.VAL.le.N-M).
                       splmode  < 0: It is assumed that the contents of
                                 X, W, M, N, and WK have not been
                                 modified since the previous invoca-
                                 tion of GCVSPL. If MD < -1, WK(4)
                                 is used as an initial estimate for
                                 the smoothing parameter p.  At the
                                 first call to GCVSPL, MD must be > 0.
                       Other values for |MD|, and inappropriate values
                       for VAL will result in an error condition, or
                       cause a default value for VAL to be selected.
                       After return from MD.ne.1, the same number of
                       degrees of freedom can be obtained, for identical
                       weight factors and knot positions, by selecting
                       |MD|=1, and by copying the value of p from WK(4)
                       into VAL. In this way, no iterative optimization
                       is required when processing other data in Y.     
   
    see gcvspl.f for more info about the two last parameter
    
    Output:
    c: the spline
    wk: work vector, see gcvspl.f
    ier: error parameter. 
        ier = 0: Normal exit 
        ier = 1: M.le.0 .or. N.lt.2*M
        ier = 2: Knot sequence is not strictly
                 increasing, or some weight
                 factor is not positive.
        ier = 3: Wrong mode parameter or value.
    
    Wrapped by C. Le Losq. with using f2py. See the howto associated for wrapping.
    
    """
    # We will do the effort in the following to keep the same name as in gcvspl.f    
    # The following stuffs are useless since gcvspl will handle that
    # by itself. However the wrapper assumes the following statement:
    # dimensions = shape(y)
    # K = dimensions[1] # number of dataset, always 1 for our case
    # N = dimensions[0] # number of independent variable
    # NY = N # Number of observation, N = NY

    wx = 1/(ese**2) # relative variance of observations
    wy = np.zeros((1))+1 # systematic errors... not used so put them to 1
    
    # Default values
    if options.get("splorder") == None:
        splorder = 2
    else:
        splorder = options.get("splorder")   
        
    if options.get("splmode") == None:
        splmode = 2
    else:
        splmode = options.get("splmode")  
        
    if options.get("NC") == None:
        NC = len(y)
    else:
        NC = options.get("NC")      

    c, wk, ier = gcvspl.gcvspl(x,y,wx,wy,splorder,splmode,VAL,NC)
    
    return c, wk, ier

def splderivative(xfull,xparse,cparse,**options):
    """
    Wrapper to the SPLDER function of gcvspl.f
    Powerful for interpolation purpose
    
    xfull (1D array) contains the entire x range where the spline has 
    to be evaluated
    xparse (1D array) contains the x values of interpolation regions   
    WARNING xparse[0] <= xfull[0] <= xparse[n] 
    cparse (1D array) is the evaluated spline coefficients returned by gcvspl
    for xparse
    
    options:    
        splineorder (integer): is the spline order, 
        default: splineorder = 2 (cubic)
        L (integer): see gcvspl.f for details, default: L = 1
        IDER: the Derivative order required, with 0.le.IDER 
        and IDER.le.2*M. If IDER.eq.0, the function
        value is returned; otherwise, the IDER-th
        derivative of the spline is returned.
        
    see the splder function in gcvspl.f for more information
    """    
   
    # Default values
    if options.get("splineorder") == None:
        splineorder = 2
    else:
        splineorder = options.get("splineorder")
        
    if options.get("L") == None:
        L = 1
    else:
        L = options.get("L")
        
    if options.get("IDER") == None:
        IDER = 0
    else:
        IDER = options.get("IDER")
        
    ycalc = gcvspl.splderv(IDER,splineorder,xfull,xparse,cparse,L)

    return ycalc



import numpy as np
from gcvspline import gcvspline, splderivative
from numpy.testing.decorators import skipif

@skipif(True, "The test may not always reach a solution at large smoothing; fix this first")
def test_sanity():
    # generating a dumb dataset for using the spline
    x = np.linspace(0, 100, 1000)
    y = 0.5 + x*30 - 0.001*x**2 + 4.5e-4 * x**3 + 1.64e-6*x**4 +500*np.cos(x/8)
    noise = np.random.normal(scale=300, size=len(y))# adding noise ,  a lot!
    y = y + noise # adding noise ,  a lot!

    # now using the spline,  we are going to store results for different smoothing
    # factors in an array y_fit
    runs = 6
    y_fit = np.zeros((np.size(x, 0), runs)) # 10 runs
    ese = np.sqrt(np.abs(y)) # we are not supposed to know the noise,  so we guess (badly) that it's something like that...
    smoothing = 0.1 # this is the interesting parameter that will allow to grow of shrink the error level,  such that we have control over the spline smooting habilities
    VAL = ese**2 # in mode 3,  see the docs and Holtring article for further details.
    for i in range(0, runs):
        c,  wk,  ier = gcvspline(x, y, smoothing*ese, VAL, splmode = 3) # gcvspl with mode 3 and smooth factor applied on data errors
        y_fit[:, i] = splderivative(x, x, c) 
        smoothing = smoothing*3 # we double the smoothing factor at each run

    # FIXME: add some assertions. 
    print(y_fit)
    # if no expection were raised the code at least ran.



from gcvspline.gcvspline2 import SmoothedNSpline, GCVSmoothedNSpline, MSESmoothedNSpline, DOFSmoothedNSpline

import numpy as np
from numpy.testing import assert_allclose, assert_raises

def test_spline_cv(): 
    def f(x):
        return 4 * x ** 3 + 10 * x ** 2 + 4 * x + 3
    x = np.linspace(-10, 10, 40)
    y = f(x)

    spl = GCVSmoothedNSpline(x, y, kind='cubic')
    y1 = spl(x)

    # since the function is cubic, cubic spl shall work perfectly.
    assert_allclose(y1, y)

def test_spline_fixed(): 
    def f(x):
        return 4 * x ** 3 + 10 * x ** 2 + 4 * x + 3
    x = np.linspace(-10, 10, 40)
    y = f(x)

    spl = SmoothedNSpline(x, y, kind='cubic', p=0.001)
    y1 = spl(x)

    # since the function is cubic, cubic spl shall work perfectly.
    assert_allclose(y1, y, rtol=1e-3)

def test_spline_mse(): 
    def f(x):
        return 4 * x ** 3 + 10 * x ** 2 + 4 * x + 3
    x = np.linspace(-10, 10, 40)
    y = f(x)

    spl = MSESmoothedNSpline(x, y, kind='cubic', variance=0.001)
    y1 = spl(x)

    # since the function is cubic, cubic spl shall work perfectly.
    assert_allclose(y1, y, rtol=1e-3)

def test_spline_dof(): 
    def f(x):
        return 4 * x ** 3 + 10 * x ** 2 + 4 * x + 3
    x = np.linspace(-10, 10, 40)
    y = f(x)

    spl = DOFSmoothedNSpline(x, y, kind='cubic', dof=3)
    y1 = spl(x)

    # since the function is cubic, cubic spl shall work perfectly.
    assert_allclose(y1, y, rtol=1e-3)

def test_spline_raise(): 
    def f(x):
        return 4 * x ** 3 + 10 * x ** 2 + 4 * x + 3
    x = np.linspace(-10, 10, 40)
    y = f(x)

    spl = GCVSmoothedNSpline(x, y, kind='cubic',  bbox=(-1, 1), ext=2)

    assert_raises(ValueError, spl, x)

def test_spline_zeros(): 
    def f(x):
        return 4 * x ** 3 + 10 * x ** 2 + 4 * x + 3
    x = np.linspace(-10, 10, 40)
    y = f(x)

    spl = GCVSmoothedNSpline(x, y, kind='cubic',  bbox=(-1, 1), ext=1)

    y1 = spl(x)
    assert_allclose(y1[x < -1], 0)
    assert_allclose(y1[x > 1], 0)

def test_spline_consts():
    def f(x):
        return 4 * x ** 3 + 10 * x ** 2 + 4 * x + 3
    x = np.linspace(-10, 10, 40)
    y = f(x)

    spl = GCVSmoothedNSpline(x, y, kind='cubic',  bbox=(-1, 1), ext=3)

    y1 = spl(x)
    assert_allclose(y1[x < -1], spl(-1))
    assert_allclose(y1[x > 1], spl(1))



# -*- coding: utf-8 -*-
"""
scipy.interpolate style wrapper for gcvspl.

"""

__author__ = "Yu Feng"
__email__ = 'rainwoodman@gmail.com'

from . import _gcvspl as gcvspl
import numpy as np

PARAMETERS= \
r"""
            w: array_like
                Relative weights of data point in spline fitting.
                A convenient choice is to use the inverted standard derivation of observations,
                  e.g. :math:`\sigma_n^{-1}` if y = y0 + n. Usually the distribution of n is not
                  known and some estimated relative weight can be used, e.g. 1.0.
                `self.wx = w ** 2`. 
            w1: array_like
                Relative weights of each dataset in spline fitting default is 1.0.
                `self.wy = w1[None, :] ** 2`. See `w`.
            kind : int, or str
                kind / order of spline, NOTE: different from the internal gcvspline parameter.
            nc  : int
                number of control points.
            bbox : tuple, or None
                bounding box of valid region for evaluation. default is None, (x[0], x[-1])
            ext : int or str
                extrapolation mode.
"""
class GCVSplineBase(object):
    SPLORDER = {'linear' : 1, 'cubic' : 2, 'quintic' : 3, 'heptic': 4, 1 : 1, 3: 2, 5 :3, 7:4}
    EXT = {'extrapolate': 0, 'zeros' : 0, 'raise' : 2, 'const' : 3}
    def __init__(self, x, y, w=1, w1=1, MD=2, VAL=1, nc=None, kind='cubic', bbox=None, ext=0):
        """ Parameters
            ----------
            %s
            MD: int
                   |MD| = 1: Prior given value for p in VAL
                             (VAL.ge.ZERO). This is the fastest
                             use of GCVSPL, since no iteration
                             is performed in p.
                   |MD| = 2: Generalized cross validation.
                   |MD| = 3: True predicted mean-squared error,
                             with prior given variance in VAL.
                   |MD| = 4: Prior given number of degrees of
                             freedom in VAL (ZERO.le.VAL.le.N-M).
                    MD  < 0: It is assumed that the contents of
                            X, W, M, N, and WK have not been
                             modified since the previous invoca-
                             tion of GCVSPL. If MD < -1, WK(4)
                             is used as an initial estimate for
                             the smoothing parameter p.  At the
                             first call to GCVSPL, MD must be > 0.
                   Other values for |MD|, and inappropriate values
                   for VAL will result in an error condition, or
                   cause a default value for VAL to be selected.
                   After return from MD.ne.1, the same number of
                   degrees of freedom can be obtained, for identical
                   weight factors and knot positions, by selecting
                   |MD|=1, and by copying the value of p from WK(4)
                   into VAL. In this way, no iterative optimization
                   is required when processing other data in Y.
            VAL: float
                parameter controls the behavior of the prior, see MD.

        """

        if not (x[:-1] < x[1:]).all():
            raise ValueError("x must be strictly increasing")

        if len(y.shape) > 2:
            raise ValueError("y must be at most two dimensional (len(x), ndataset)")

        if len(y.shape) == 1:
            y = y.reshape(-1, 1)

        if bbox is None:
            bbox = (x[0], x[-1])

        wx = np.ones(y.shape[0])
        wx[...] = w ** 2
        wy = np.ones(y.shape[1])
        wy[...] = w1 ** 2

        if nc is None: nc = len(x)

        splorder = self.SPLORDER.get(kind)

        if splorder not in (1, 2, 3, 4):
            raise ValueError("splorder must be 1, 2, 3, 4.")

        if MD not in (1, 2, 3, 4):
            raise ValueError("MD must be 1, 2, 3, 4")

        ext = self.EXT.get(ext, ext)
        if ext not in (0, 1, 2, 3):
            raise ValueError("exts must be 1, 2, 3, 4")

        c, wk, ier = gcvspl.gcvspl(x, y, wx, wy, splorder, MD, VAL, nc)
        if ier == 1:
            raise RuntimeError(" M <= 0 or N < 2 * M ")
        elif ier == 2:
            raise RuntimeError("Knot sequence is not strictly increasing"
                    + " or some weight factor is not positive.")
        elif ier == 3:
            raise RuntimeError("Wrong mode parameter or value")

        self.x = x
        self.c = c
        self.wk = wk
        self.wy = wy[None, :]
        self.wx = wx[:, None]
        self.splorder = splorder
        self.L = 1
        self.bbox = np.array(bbox, dtype='f8')
        self.ext = ext

    @property
    def gcv_value(self):
        """ Generalized Cross Validation value """
        return self.wk[0]

    @property
    def msr(self):
        r""" Mean Squared Residual :math:`\sigma^2 (y - \hat{f}(x))` """
        return self.wk[1]

    @property
    def dof(self):
        """ Estimated DOF """
        return self.wk[2]

    @property
    def p(self):
        """ Smoothing Parameter p. p degenerated with the weights. """
        return self.wk[3]

    @property
    def mse(self):
        """ Estimated MSE. mse + msr = variance_metric. """
        return self.wk[4]

    @property
    def variance_metric(self):
        """ Gauss-Markov error variance metric. """
        return self.wk[5]

    @property
    def variance(self):
        """ Estimated variance of each data entry.
            Scaling the weights (after converting to variance) by variance_metric. """
        return self.wk[5] / (self.wy * self.wx)

    def __call__(self, x, nu=0, ext=None):
        x = np.array(x)
        oldshape = x.shape
        x = np.ravel(x)
        if ext is None:
            ext = self.ext
        if ext != 0:
            bmask1 = (x < self.bbox[0])
            bmask2 = (x > self.bbox[1])
        if ext == 3:
            x = np.copy(x)
            x[bmask1] = self.bbox[0]
            x[bmask2] = self.bbox[1]
        elif ext == 2:
            if bmask1.any() or bmask2.any():
                raise ValueError("some values are out of bound")

        y = gcvspl.splderv(nu, self.splorder, x, self.x, self.c, self.L)

        if ext == 1:
            y[bmask1 | bmask2] = 0

        return y.reshape(oldshape)

class SmoothedNSpline(GCVSplineBase):
    def __init__(self, x, y, p, w=1, w1=1, nc=None, kind='cubic', bbox=None, ext=0):
        """ Natural Spline fitting with fixed smoothing

            Parameters
            ----------
            %s
            p : float
                smoothing parameter. The parameter p is degenerate with weights.
        """
        GCVSplineBase.__init__(self, x=x, y=y, w=w, w1=w1, nc=nc, kind=kind, bbox=bbox, ext=ext, VAL=p, MD=1)

class GCVSmoothedNSpline(GCVSplineBase):
    def __init__(self, x, y, w=1, w1=1, nc=None, kind='cubic', bbox=None, ext=0):
        """ Natural Spline fitting with smoothing selected by cross validation.

            Parameters
            ----------
            %s
        """
        GCVSplineBase.__init__(self, x=x, y=y, w=w, w1=w1, nc=nc, kind=kind, bbox=bbox, ext=ext, VAL=0, MD=2)

class MSESmoothedNSpline(GCVSplineBase):
    def __init__(self, x, y, w=1, w1=1, nc=None, kind='cubic', bbox=None, ext=0, variance_metric=1.0):
        """ Natural Spline fitting with known variance_metric, minimizing the MSE.

            Parameters
            ----------
            %s
            variance_metric: float
                smoothing terms of variance metric. The default value of 1.0 is `appropriate`
                when `w` and `w1` are already scaled as the inverted std error of the noise, 
                and when the underlying model is a spline (mse == 0).
                Larger number produces smoother results. The sets the same named property of
                GCVSmoothedNSpline.
        """
        GCVSplineBase.__init__(self, x=x, y=y, w=w, w1=w1, nc=nc, kind=kind, bbox=bbox, ext=ext, VAL=variance_metric, MD=3)

class DOFSmoothedNSpline(GCVSplineBase):
    def __init__(self, x, y, dof, w=1, w1=1, nc=None, kind='cubic', bbox=None, ext=0):
        """ Nautural Spline fitting with known degrees of freedom.

            Parameters
            ----------
            %s
            dof : float
                desired effective degrees of freedom of smoothing. Large number means more smoothing.
        """
        GCVSplineBase.__init__(self, x=x, y=y, w=w, w1=w1, nc=nc, kind=kind, bbox=bbox, ext=ext, VAL=dof, MD=4)

def adddocstring(klass):
    klass.__init__.__doc__ = klass.__init__.__doc__ % PARAMETERS.strip()

adddocstring(GCVSplineBase)
adddocstring(SmoothedNSpline)
adddocstring(GCVSmoothedNSpline)
adddocstring(MSESmoothedNSpline)
adddocstring(DOFSmoothedNSpline)




from numpy.distutils.core import setup
from numpy.distutils.extension import Extension

extension = Extension("gcvspline._gcvspl", ["gcvspline/gcvspl.f"])

setup(name='gcvspline',
      version='0.3',
      description='A Python wrapper for gcvspl.f, a FORTRAN package for generalized, cross-validatory spline smoothing and differentiation.',
      url='https://github.com/charlesll/gcvspline',
      author='Charles Le Losq and Yu Feng',
      author_email='charles.lelosq@anu.edu.au',
      license='GNU-GPLv3',
      packages=['gcvspline', 'gcvspline.tests'],
      ext_modules=[extension],
      zip_safe=False)



import numpy as np
from matplotlib import pyplot as plt
from gcvspline import gcvspline, splderivative

# generating a dumb dataset for using the spline
x = np.linspace(0,100,1000)
y = 0.5 + x*30 - 0.001*x**2 + 4.5e-4 * x**3 + 1.64e-6*x**4 +500*np.cos(x/8)
noise =np.random.normal(scale=300,size=len(y))# adding noise , a lot!
y = y+ noise # adding noise , a lot!

# now using the spline, we are going to store results for different smoothing
# factors in an array y_fit
runs = 6
y_fit = np.zeros((np.size(x,0),runs)) # 10 runs
ese = np.sqrt(np.abs(y)) # we are not supposed to know the noise, so we guess (badly) that it's something like that...
smoothing = 0.1 # this is the interesting parameter that will allow to grow of shrink the error level, such that we have control over the spline smooting habilities
VAL = ese**2 # in mode 3, see the docs and Holtring article for further details.
for i in range(0,runs):
    c, wk, ier = gcvspline(x,y,smoothing*ese,VAL,splmode = 3) # gcvspl with mode 3 and smooth factor applied on data errors
    y_fit[:,i] = splderivative(x,x,c) 
    smoothing = smoothing*3 # we double the smoothing factor at each run

plt.figure()
plt.plot(x,y,'k.')
smoothing = 0.01
for i in range(0,runs): 
    if i > 0: 
        thickness = 3.0 
    else: 
        thickness = 0.5
    plt.plot(x,y_fit[:,i],'--',linewidth=thickness,color=[0+float(i)/float(runs-1.0),0,0],label=str(round(smoothing)))
    smoothing = smoothing*3 # we double the smoothing factor at each run

plt.xlabel('X values')
plt.ylabel('Y values')
plt.title('A simple example')
plt.legend(loc=2)
